#include <Eigen/Core>
#include <me530646_lab1/lab1.h>
#include <ur5/utilities.h>
#include <math.h>
#include <iostream>

#include <stdio.h>     
#define PI 3.1415926

/**
 * Returns the 3x3 rotation matrix that represents a rotation about 
 * the x axis (roll) of phi radians.
 */
Eigen::Matrix3f rollr(double phi){
	Eigen::Matrix3f rollx(3,3);
	rollx<< 1,0,0,
		0,cos(phi),-sin(phi),
		0,sin(phi),cos(phi);
	return rollx;
	//done
}

/**
 * Returns the 3x3 rotation matrix that represents a rotation about 
 * the y axis (pitch) of theta radians.
 */
Eigen::Matrix3f pitchr(double theta){
	Eigen::Matrix3f pitchy(3,3);
	pitchy<< cos(theta),0,sin(theta),
		0,1,0,
		-sin(theta),0,cos(theta);
	return pitchy;
	//done		 
}

/**
 * Returns the 3x3 rotation matrix that represents a rotation about 
 * the z axis (yaw) of psi radians.
 */
Eigen::Matrix3f yawr(double psi){
	Eigen::Matrix3f yawz(3,3);
	yawz<< cos(psi),-sin(psi),0,
		sin(psi),cos(psi),0,
		0,0,1;
	return yawz;
	//done
}

/**
 * Returns the 3x3 rotation matrix that represents successive 
 * roll, pitch, and yaw rotations.
 */
Eigen::Matrix3f rpyr(double roll, double pitch, double yaw){
	Eigen::Matrix3f rpyxyz(3,3);
	rpyxyz=yawr(yaw)*pitchr(pitch)*rollr(roll);
	return rpyxyz;
	//done  Assume it rotates about extrinsic axis           
}

/**
 * Returns the roll, pitch, and yaw from the given 3x3 rotation matrix.
 */
Eigen::Vector3f rpyrinv(Eigen::Matrix3f r){
	Eigen::Vector3f angles(3);
	float rollx=0;
	float pitchy=0;
	float yawz=0;
	pitchy=asin(-r(2,0));  //-pi/2<=pitchy<=pi/2      
	if(r(0,0)>0.000001||r(0,0)<-0.000001||r(1,0)>0.000001||r(1,0)<-0.000001){  //cos(pitchy)!=0 
				   //Since PI is an approximate number use 0.000001 instead of 0 
		yawz=atan2(r(1,0),r(0,0));  //-pi/2<=yawz<=pi/2     
		rollx=atan2(r(2,1),r(2,2));  //-pi/2<=rollx<=pi/2   
                //By analysing the matrix r, there's no way to distinguish between  
                //(rollx pitchy yawz) and ( rollx+PI  PI-pitchy  yawz+PI) 
                //when 0<pitchy<PI/2 (or -PI/2<pitchy<0) since they yield the same r          
		
	}

	else{
		if(fabs(r(2,0)-(-1))<0.000001)  //sin(pitchy)==1
			rollx=atan2(r(0,1),r(1,1))+yawz;
		else  //sin(pitchy)==-1
			rollx=atan2(-r(0,1),r(1,1))-yawz;	
		std::cout<<"When pitch angle equals (+/-)pi/2 The other two angles cannot be calculated"<<std::endl;
	}

	angles(0)=rollx;
	angles(1)=pitchy;
	angles(2)=yawz;

	return angles;
		
	//done  Assume r is the conventional extrinsic rotation matrix calculated in rpyr      
}

/**
 * Returns the 4x4 homogeneous transformation that represents a 
 * translation of (x,y,z) and a rotation of (roll,pitch,yaw).
 */
Eigen::Matrix4f xf(double x, double y, double z, double roll, double pitch, double yaw){
	Eigen::Matrix4f homotransf(4,4);
	Eigen::Matrix3f rpyxyz(3,3);
	rpyxyz=rpyr(roll, pitch, yaw);
	homotransf<<rpyxyz(0,0),rpyxyz(0,1),rpyxyz(0,2),x,
		    rpyxyz(1,0),rpyxyz(1,1),rpyxyz(1,2),y,
		    rpyxyz(2,0),rpyxyz(2,1),rpyxyz(2,2),z,
		    0,0,0,1;
	return homotransf;   
	
	//done   
}

/**
 * Returns the matrix inverse to the given homogeneous transformation.
 */
Eigen::Matrix4f finv(Eigen::Matrix4f f){
	Eigen::Matrix4f finv(4,4);
	Eigen::Matrix3f rot(3,3);
	rot=f.block<3,3>(0,0);
	finv.block<3,3>(0,0)=rot.transpose();
	finv.block<3,1>(0,3)=-rot.transpose()*f.block<3,1>(0,3);
	finv.row(3)=f.row(3);
	return finv;
	
	//done  the inverse transformation      
}

/**
 * Animates a roll-pitch-yaw rotation about the world
 * coordinate frame
 */
void XYZFixedPlot(RvizPlotter &plotter, double roll, double pitch, double yaw){
	Eigen::Matrix3f rot(3,3);
	Eigen::Matrix4f homotransf = Eigen::MatrixXf::Identity(4,4);
	rot=rollr(roll)*pitchr(pitch)*yawr(yaw);  //the rotation matrix about fixed axis in the order
                                                  //of Z Y X       
	homotransf.block<3,3>(0,0)=rot;      
	plotter.plotf(homotransf,"RotFix");

	//done
}

/**
 * Animates a roll-pitch-yaw rotation about the body-fixed
 * coordinate frame
 */
void ZYXRelativePlot(RvizPlotter &plotter, double roll, double pitch, double yaw){
	Eigen::Matrix3f rot(3,3);
	Eigen::Matrix4f homotransf = Eigen::MatrixXf::Identity(4,4);
	rot=rollr(roll)*pitchr(pitch)*yawr(yaw);  //the rotation matrix about relative axis in the 
                                                  //X Y Z   
	homotransf.block<3,3>(0,0)=rot;
	plotter.plotf(homotransf,"RotRelative");

	//done   
}
