#include <Eigen/Dense>
#include <unsupported/Eigen/MatrixFunctions>
#include <me530646_lab1/lab1.h>
#include <Eigen/Eigenvalues>

/**
 * Returns the 3x3 skew-symmetric matrix that corresponds to vector e.
 */
Eigen::Matrix3f skew3(Eigen::Vector3f e){
    Eigen::Matrix3f ehat(3,3);
    ehat<<0,-e(2),e(1),
	  e(2),0,-e(0),
	  -e(1),e(0),0;
    return ehat;
	//Done
}

/**
 * Returns the 3x3 rotation matrix that corresponds to a rotation of 
 * |e| radians about vector e.
 */
Eigen::Matrix3f expr(Eigen::Vector3f e){
    Eigen::Matrix3f exprot(3,3);  //the rotation matrix represented by the exponential of ehat     
    Eigen::Matrix3f ehat(3,3);
    //Eigen::Matrix3f rot(3,3);  //the matrix representing the axis and the angle     
    float enorm;     
    ehat=skew3(e);
    enorm=e.norm();
    //rot=ehat*enorm;
    //exprot=rot.exp();    
    exprot=ehat.exp(); 
    return exprot;       
	//Done     
}

/**
 * Returns the (x,y,z) translation and (roll,pitch,yaw) rotations
 * of the given homogeneous transformation.
 */
Eigen::VectorXf xfinv(Eigen::Matrix4f xf){
    Eigen::VectorXf rpyxyz(6);     
    Eigen::Matrix3f rpyxyzr(3,3);     //the rotation matrix of rpy          
    rpyxyzr=xf.block<3,3>(0,0);
    rpyxyz(3)=rpyrinv(rpyxyzr)(0);
    rpyxyz(4)=rpyrinv(rpyxyzr)(1);
    rpyxyz(5)=rpyrinv(rpyxyzr)(2);
    rpyxyz(0)=xf(0,3);
    rpyxyz(1)=xf(1,3);
    rpyxyz(2)=xf(2,3);     
	//Done     
    return rpyxyz;
}
     
     
         

