#include <Eigen/Core>
#include <ur5/utilities.h>
#include <unsupported/Eigen/MatrixFunctions>
#include <me530646_lab1/lab1.h>
#include <me530646_lab2/lab2.h>
#include <Eigen/Eigenvalues>      

#include <stdio.h>             
#define PI 3.1415926     
using std::endl;  

int main(int argc, char **argv){
	//Initializing ROS
    ros::init(argc, argv, "lab1_main");
    ros::NodeHandle node;
    //Creating an RvizPlotter
	RvizPlotter p = RvizPlotter(node);
	
	/*****Examples of how to use necessary functions*****
	//Calculating the norm of vector v
	Eigen::Vector3f v = Eigen::MatrixXf::Random(3,1);
	double normV = v.norm();
	//Calculating the matrix expontential of matrix m
	Eigen::Matrix4f m = Eigen::MatrixXf::Random(4,4);
	Eigen::Matrix4f expM = m.exp();
	//Calculating the eigenvalues and eigenvectors of m
	Eigen::EigenSolver<Eigen::MatrixXf> es(m);
	//VectorXcf MatrixXcf are used because ,there may 
	//be complex values for eigenvalues and eigenvectors
	Eigen::VectorXcf evals = es.eigenvalues();
	Eigen::MatrixXcf evecs = es.eigenvectors();
	****************************************************/





    Eigen::Vector3f vabt(3);    //an arbitrary 3*1 vector     
    double normv;     
    Eigen::Vector3f vunit(3);    
    Eigen::Matrix3f vhat(3,3);    
    Eigen::Vector3f vzero(3);  
    Eigen::Matrix3f expvhat(3,3);

    vabt<<1,3,9;
    std::cout<<"Question2 Choose the vector vabt to be "<<'\n'<<vabt<<'\n'<<std::endl;     
    
    //Question 2(a) Calculating the norm of vector v  
    normv=vabt.norm();   
    std::cout<<"the norm of v is: "<<'\n'<<normv<<'\n'<<std::endl;

    //2(b) Calculating the unit vector of vector v
    vunit=vabt/normv;     
    std::cout<<"the unit vector of v is: "<<'\n'<<vunit<<'\n'<<std::endl;

    //2(c) Calculating the skew symmetric matrix vhat    
    vhat=skew3(vabt);     
    std::cout<<"the skew symmetric matrix of v is: "<<'\n'<<vhat<<'\n'<<std::endl;

    //2(d) Show that vabt is an element of the kernal of vhat    
    vzero<<0,0,0;  
    if(vhat*vabt==vzero)
	std::cout<<"v is an element of the kernal of its skew symmetric matrix"<<'\n'<<std::endl;  
    else
	std::cout<<"v is not an element of the kernal of its skew symmetric matrix"<<'\n'<<std::endl;
    std::cout<<"vhat*vabt = "<<vhat*vabt<<'\n'<<std::endl;

    //2(e) Calculating the eigenvalues and eigenvectors of vhat    
    Eigen::EigenSolver<Eigen::MatrixXf> es(vhat);
    Eigen::VectorXcf vabtevalues=es.eigenvalues();
    Eigen::MatrixXcf vabtevectors=es.eigenvectors();     
    std::cout<<"the eigenvalues of v are: "<<'\n'<<vabtevalues<<'\n'<<endl;
    std::cout<<"the eigenvectors of v are: "<<'\n'<<vabtevectors<<'\n'<<endl;     
    //Their relationship is the eigenvector of the eigenvalue 0 is the scale of vector v        
    //In other words vhat*v==0       

    //2(f) Calculating exp(vhat)    
    expvhat=vhat.exp();     
    std::cout<<"exp(vhat) is: "<<'\n'<<expvhat<<'\n'<<std::endl;     

    //2(g) Show that v=exp(vhat)*v 
    Eigen::Vector3f productvhatv(3); 
    productvhatv=expvhat*vabt-vabt;    
    if(productvhatv(0)<=0.000001&&productvhatv(1)<=0.000001&&productvhatv(2)<=0.000001)     
	std::cout<<"v==exp(vhat)*v"<<'\n'<<std::endl;
    else
	std::cout<<"v!=exp(vhat)*v"<<'\n'<<std::endl;

    //2(h) Calculating the eigenvalues and eigenvectors of exp(vhat)     
    Eigen::EigenSolver<Eigen::MatrixXf> es1(expvhat);     
    Eigen::VectorXcf expvhatevalues = es1.eigenvalues();
    Eigen::MatrixXcf expvhatevectors = es1.eigenvectors();     
    std::cout<<"the eigenvalues of expvhat is: "<<'\n'<<expvhatevalues<<'\n'<<std::endl;     
    std::cout<<"the eigenvectors of expvhat are: "<<'\n'<<expvhatevectors<<'\n'<<std::endl;  
    //Their relationship is the eigenvector of eigenvalue 1 is the scale of vector v       
    //In other words exp(vhat)*v=1*v
               


    //Question 3 Use expr() to generate rotation matrices with various angles and fixed axes         
    //Use plotf() to plot the original and rotated frames in rviz   

    //let the fixed rotation axes be raxis=i*(0.1,0.3,0.9)     
    //where i*norm(0.1,0.3,0.9) indicates different radians  
    //and experiment with 300 different radians     
    int n=300;   
    int i=1;     
    Eigen::Vector3f raxisinitial(3);
    Eigen::Vector3f raxis(3);
    Eigen::Matrix3f exprot(3,3);
    Eigen::Matrix4f hexprot = Eigen::MatrixXf::Identity(4,4);
    raxisinitial<<0.01,0.03,0.09;   
    for (i=1;i<=n;i++){
	raxis=i*raxisinitial;
	exprot=expr(raxis);   //exprot are the rotation matrices for different rotation radians    
	hexprot.block<3,3>(0,0)=exprot; 
	p.plotf(hexprot,"rot");     
    }     





    //Question 4(a) xf(x,y,z,r,p,y)=xf(x,y,z,r + 2 PI,p,y) ∀ x, y, z, r, p, y ∈ R.
    Eigen::VectorXf rotparam = Eigen::VectorXf::Random(6);      //the random parameters (x,y,z,r,p,y)  
    Eigen::VectorXf rpyxyz(6); 
    Eigen::Matrix4f homotransf1(4,4);     
    Eigen::Matrix4f homotransf2(4,4);
    homotransf1=xf(rotparam(0),rotparam(1), rotparam(2),rotparam(3),rotparam(4),rotparam(5));
    homotransf2=xf(rotparam(0),rotparam(1), rotparam(2),rotparam(3)+2*PI,rotparam(4),rotparam(5));  
    int j=0;  
    int equalyes=1; 
    for (i=0;i<4;i++){
	for(j=0;j<4;j++){
	if(homotransf1(i,j)-homotransf2(i,j)<-0.000001||homotransf1(i,j)-homotransf2(i,j)>0.000001)   
	equalyes=0;
	}
    }
    if(equalyes==1)
	std::cout<<"Question4 xf(x,y,z,r,p,y)=xf(x,y,z,r+2*PI,p,y) demonstrated"<<'\n'<<std::endl;    
    else
	std::cout<<"xf(x,y,z,r,p,y)!=xf(x,y,z,r+2*PI,p,y) in some cases"<<'\n'<<std::endl;   

    //Question 4(b) Answer (more details in the lab2.pdf)         
    //Because xfinv(xf(x,y,z,r+2*PI,p,y))==(x,y,z,r,p,y) not (x,y,z,r+2*PI,p,y), similarly to p and y  
    //And because when H(0,0)==H(1,0)==0, can only get p==PI/2, but r and y can't be got     

    //Question 4(c)      
    rotparam<<0, 0, 0, 4*PI, 0, 0;     
    homotransf1=xf(rotparam(0),rotparam(1), rotparam(2),rotparam(3),rotparam(4),rotparam(5));
    rpyxyz=xfinv(homotransf1);     
    std::cout<<"the inverse rotation parameters: "<<'\n'<<rpyxyz<<'\n'<<std::endl;   

    //Question 4(d)     
    rotparam = Eigen::VectorXf::Random(6);          
    for(i=3;i<=5;i++)
	rotparam(i)=rotparam(i)-floor(rotparam(i)/(2*PI))*2*PI;
    if(rotparam(4)==PI/2)
	rotparam(4)=rotparam(4)+0.5;     
    homotransf1=xf(rotparam(0),rotparam(1), rotparam(2),rotparam(3),rotparam(4),rotparam(5));
    rpyxyz=xfinv(homotransf1);   
    homotransf2=xf(rpyxyz(0),rpyxyz(1),rpyxyz(2),rpyxyz(3),rpyxyz(4),rpyxyz(5));     
    std::cout<<"the transformation parameters: "<<'\n'<<rotparam<<'\n'<<std::endl; 
    std::cout<<"the original homogeneous transformation: "<<'\n'<<homotransf1<<'\n'<<std::endl;
    std::cout<<"the inverse homogeneous representation: "<<'\n'<<homotransf2<<'\n'<<std::endl;     


 



    //Question 5(b)      
    float q1, q2, q3;     
    Eigen::Vector3f point1,point2,point3,point4,point5;  //to generate random vectors in frame3     
   
    q1=5.0*PI/4;     
    q2=5.0*PI/4;   
    q3=5.0*PI/4;      
    point1<<0.1,0.2,0.3;     
    point2<<0.5,0.3,0.9;
    point3<<0.7,0.6,0.03;   
    point4<<0.22,0.75,0.15;
    point5<<0.52,0.01,0.96;  
    Eigen::Matrix4f homotransf01(4,4);     
    Eigen::Matrix4f homotransf12(4,4);
    Eigen::Matrix4f homotransf23(4,4);     
    Eigen::Matrix4f homotransf03(4,4);
    homotransf01=xf(q1,0,0,0,0,0);
    homotransf12=xf(0,0,0,q2,0,0);
    homotransf23=xf(0,0,q3,0,0,0);
    p.plotf(homotransf01, "frame1");
    p.plotf(homotransf12, "frame1", "frame2");   
    p.plotf(homotransf23, "frame2", "frame3");    
    p.plotv("frame3", point1, point2);
    p.plotv("frame3", point3, point4);     
    p.plotv("frame3", point4, point5);

    homotransf03=getTransformation("map","frame3");     
    std::cout<<"Question5 transformation at qi==5*PI/4: "<<'\n'<<homotransf03<<'\n'<<std::endl;      
  
    
    
    
    
    return 0;    	

}
