#include <Eigen/Core>
#include <ur5/utilities.h>
#include <math.h>
#include <me530646_lab1/lab1.h>
#include <me530646_lab2/lab2.h>
#include <me530646_lab3/lab3.h>
#include <me530646_lab4/lab4.h>

#define PI M_PI
#define approxZero 0.00001


#define Notangle 500     

double alpha[] = {PI/2,0,0,PI/2,-PI/2,0};
double a[] = {0,-0.425,-0.392,0,0,0};
double d[] = {0.0892,0,0,0.1093,0.09475,0.0825};

Eigen::MatrixXf J(double q[6])
{
    Eigen::MatrixXf Jcb=Eigen::MatrixXf::Random(6,6);
    Eigen::Matrix4f homotransfbg(4,4);
    homotransfbg = UR5::fwd(q);     
    int i;
    Eigen::Vector3f e1(3,1);
    Eigen::Vector3f e2(3,1);
    Eigen::Vector3f e3(3,1);
    Eigen::Matrix4f homotransf[6];
    Eigen::Matrix4f homotransfbi[6];
    Eigen::Matrix3f rot[6];             
    Eigen::Matrix3f I3=Eigen::MatrixXf::Identity(3,3);    
    e1<<1,0,0;
    e2<<0,1,0;
    e3<<0,0,1; 
    for (i=0;i<6;i++){                            
	homotransf[i]=UR5::dhf(alpha[i], a[i], d[i], q[i]);
	if(i>0)
	    homotransfbi[i]=homotransfbi[i-1]*homotransf[i];
	else
	    homotransfbi[0]=homotransf[0];	
    }
    for (i=0;i<6;i++){     //to calculate Jw 
	rot[i]=homotransfbi[i].block<3,3>(0,0);
	if(i==0)
	    Jcb.block<3,1>(3,0)=I3*e3;
	else
	    Jcb.block<3,1>(3,i)=rot[i-1]*e3;  
    }

    Jcb.block<3,1>(0,5)<<0,0,0;     //to calculate Jv     
    Jcb.block<3,1>(0,4)=(-d[4])*rot[4]*(-e1)+Jcb.block<3,1>(0,5);
    Jcb.block<3,1>(0,3)=d[3]*rot[3]*(-e1)+Jcb.block<3,1>(0,4);     
    Jcb.block<3,1>(0,2)=a[2]*rot[2]*e2+Jcb.block<3,1>(0,3);
    Jcb.block<3,1>(0,1)=a[1]*rot[1]*e2+Jcb.block<3,1>(0,2);
    Jcb.block<3,1>(0,0)=d[0]*rot[0]*(-e1)+Jcb.block<3,1>(0,1);
    
    return Jcb;
	
//Done     
}




int inverse(Eigen::Matrix4f H0_6, double **q, double q6Des){
  int numSols = 0;

  double parccos, parcsin;     //the parameter of the arccos and arcsin function     
  double patan2y, patan2x;     //the parameters of the atan2 function     
  Eigen::Vector3f e1(1,0,0);
  Eigen::Vector3f e2(0,1,0);
  Eigen::Vector3f e3(0,0,1);
  int n;

  /*==========================================
  /       Solving for q1
  /=========================================*/
  double q1[2];
  Eigen::Vector4f P05(4,1);  //P05 is the vector poiting to the origin of frame 5 in the base frame     
  Eigen::Vector4f P65(4,1);  //P65 is the coordinate of the origin of frame 5 in frame 6   
  P65<<0,0,-d[5],1;     
  P05=H0_6*P65;
  P05(3)=0;
  double ang1,ang2[2];
  ang1=atan2(P05(1),P05(0));
  parccos=d[3]/double(sqrt(P05(0)*P05(0)+P05(1)*P05(1)));

  if(parccos<=1&&parccos>=-1){
    ang2[0]=acos(parccos);
    ang2[1]=-ang2[0];   
    q1[0]=ang1+ang2[0]+PI/2.0;
    q1[1]=ang1+ang2[1]+PI/2.0;
    while(q1[0]<0)
	q1[0]=q1[0]+2*PI;
    while(q1[0]>2*PI)
	q1[0]=q1[0]-2*PI;  
    if(q1[0]<=approxZero&&q1[0]>=-approxZero)
	q1[0]=0;
  
    while(q1[1]<0)
	q1[1]=q1[1]+2*PI;
    while(q1[1]>2*PI)
	q1[1]=q1[1]-2*PI;  
    if(q1[1]<=approxZero&&q1[1]>=-approxZero)
	q1[1]=0;
   }

  else{
    q1[0]=Notangle;
    q1[1]=Notangle;
    
  }

 

  /*==========================================
  /       Solving for q5
  /=========================================*/
  double q5[2][2];
  Eigen::Vector4f P06(4,1);
  double P16z[2];     
  P06=H0_6.col(3);
  P06(3)=0;
  P16z[0]=P06(0)*sin(q1[0])-P06(1)*cos(q1[0]);
  parccos=(P16z[0]-d[3])/d[5];
  if(parccos<=1&&parccos>=-1){
    q5[0][0]=acos(parccos);
    q5[0][1]=-q5[0][0];
    for (n=0;n<2;n++){
	while(q5[0][n]<0)
	  q5[0][n]=q5[0][n]+2*PI;
        while(q5[0][n]>2*PI)
	  q5[0][n]=q5[0][n]-2*PI;  
        if(q5[0][n]<=approxZero&&q5[0][n]>=-approxZero)
	  q5[0][n]=0;
    }
  }
  else{
    q5[0][0]=Notangle;
    q5[0][1]=Notangle;
  }
    
  P16z[1]=P06(0)*sin(q1[1])-P06(1)*cos(q1[1]);
  parccos=(P16z[1]-d[3])/d[5];
  if(parccos<=1&&parccos>=-1){
    q5[1][0]=acos(parccos);
    q5[1][1]=-q5[1][0];
    for (n=0;n<2;n++){
	while(q5[1][n]<0)
	  q5[1][n]=q5[1][n]+2*PI;
        while(q5[1][n]>2*PI)
	  q5[1][n]=q5[1][n]-2*PI;  
        if(q5[1][n]<=approxZero&&q5[1][n]>=-approxZero)
	  q5[1][n]=0;
    }
  }
  else{
    q5[1][0]=Notangle;
    q5[1][1]=Notangle;
  }
   


  /*==========================================
  /       Solving for q6,q2-q4
  /=========================================*/
  double q6[2][2];
  double q3[2][2][2];
  double q2[2][2][2];     
  double q4[2][2][2];
  
  Eigen::Matrix4f homotransfb1[2];
  Eigen::Matrix4f homotransf61[2];     //the transformation between frame 6 and 1     
  Eigen::Matrix3f rot61[2];
  Eigen::Matrix4f homotransf14[2][2];
  Eigen::Matrix4f homotransf45[2][2];
  Eigen::Matrix4f homotransf56[2][2];   
  Eigen::Vector4f P13[2][2];
  Eigen::Vector4f P43(4,1);     
  Eigen::Matrix4f homotransf12[2][2][2];
  Eigen::Matrix4f homotransf23[2][2][2];
  Eigen::Matrix4f homotransf34[2][2][2];

  
  //For each solution to q1
  for(int i = 0; i < 2; i++)
  {
    if(q1[i]==Notangle)
	break;
    homotransfb1[i]=UR5::dhf(PI/2, 0, d[0], q1[i]);
    homotransf61[i]=finv(finv(homotransfb1[i])*H0_6);
    rot61[i]=homotransf61[i].block<3,3>(0,0);

    //For each solution to q5
    for(int j = 0; j < 2; j++)
    {
      if(q5[i][j]==Notangle)
	break;
      if(sin(q5[i][j])!=0){
	patan2y=((rot61[i]*e3).dot(-e2))/sin(q5[i][j]);
        patan2x=((rot61[i]*e3).dot(e1))/sin(q5[i][j]);
        if(patan2y!=0||patan2x!=0){
          q6[i][j]=atan2(patan2y,patan2x);
          while(q6[i][j]<0)
	      q6[i][j]=q6[i][j]+2*PI;
          while(q6[i][j]>2*PI)
	      q6[i][j]=q6[i][j]-2*PI;  
          if(q6[i][j]<=approxZero&&q6[i][j]>=-approxZero)
	      q6[i][j]=0;
        }   
      }
      else
	q6[i][j]=PI/3;     //arbitrarilly choose a q6     


      homotransf45[i][j]=UR5::dhf(alpha[4], a[4], d[4], q5[i][j]);
      homotransf56[i][j]=UR5::dhf(alpha[5], a[5], d[5], q6[i][j]);
      homotransf14[i][j]=finv(homotransfb1[i])*H0_6*finv(homotransf45[i][j]*homotransf56[i][j]);
      P43<<0,-d[3],0,1;
      P13[i][j]=homotransf14[i][j]*P43;
      P13[i][j](3)=0;
      parccos=(P13[i][j].dot(P13[i][j])-a[1]*a[1]-a[2]*a[2])/(2*a[1]*a[2]);
      if(parccos<=1&&parccos>=-1){
	q3[i][j][0]=acos(parccos);
	q3[i][j][1]=-q3[i][j][0];
	for (n=0;n<2;n++){
	  while(q3[i][j][n]<0)
	    q3[i][j][n]=q3[i][j][n]+2*PI;
          while(q3[i][j][n]>2*PI)
	    q3[i][j][n]=q3[i][j][n]-2*PI;  
          if(q3[i][j][n]<=approxZero&&q3[i][j][n]>=-approxZero)
	    q3[i][j][n]=0;     
	}
      }
      else{
	q3[i][j][0]=Notangle;
	q3[i][j][1]=Notangle;
      }
      
      
      patan2y=P13[i][j](1);
      patan2x=-P13[i][j](0);
      parcsin=(a[2])*sin(q3[i][j][0])/sqrt(P13[i][j].dot(P13[i][j]));  //it is -a[2] in the ik file     
      if(parcsin<=1&&parcsin>=-1){
        if(patan2y!=0||patan2x!=0){
	  q2[i][j][0]=-atan2(patan2y,patan2x)+asin(parcsin);
	  q2[i][j][1]=-atan2(patan2y,patan2x)+asin(-parcsin);
	  for (n=0;n<2;n++){
	    while(q2[i][j][n]<0)
	      q2[i][j][n]=q2[i][j][n]+2*PI;
            while(q2[i][j][n]>2*PI)
	      q2[i][j][n]=q2[i][j][n]-2*PI;  
            if(q2[i][j][n]<=approxZero&&q2[i][j][n]>=-approxZero)
	      q2[i][j][n]=0;     
	  }
	}
	else{  
	  q2[i][j][0]=PI/3;     //arbitrarilly choose a q2 but since a2!=a3 it would not be got        
	  q2[i][j][1]=PI/3;
	}
      }
      else{
	q2[i][j][0]=Notangle;
        q2[i][j][1]=Notangle;
      }



      //For each solution to q2/q3
      for(int k = 0; k < 2; k++)
      {   
	if(q3[i][j][k]==Notangle||q2[i][j][k]==Notangle)
	  break;     
	homotransf12[i][j][k]=UR5::dhf(alpha[1], a[1], d[1], q2[i][j][k]);
	homotransf23[i][j][k]=UR5::dhf(alpha[2], a[2], d[2], q3[i][j][k]);
	homotransf34[i][j][k]=(finv(homotransf12[i][j][k]*homotransf23[i][j][k]))*homotransf14[i][j];
	patan2y=homotransf34[i][j][k](1,0);
	patan2x=homotransf34[i][j][k](0,0);
	if(patan2y!=0||patan2x!=0)
	  q4[i][j][k]=atan2(patan2y,patan2x);
	else
	  q4[i][j][k]=Notangle;     
	while(q4[i][j][k]<0)
	  q4[i][j][k]=q4[i][j][k]+2*PI;
        while(q4[i][j][k]>2*PI)
	  q4[i][j][k]=q4[i][j][k]-2*PI;  
        if(q4[i][j][k]<=approxZero&&q3[i][j][n]>=-approxZero)
	  q4[i][j][k]=0; 


	
        if(q4[i][j][k]==Notangle)
	  continue;    
	else{  
	  q[numSols][0] = q1[i]; //...
	  q[numSols][1] = q2[i][j][k];
	  q[numSols][2] = q3[i][j][k];
	  q[numSols][3] = q4[i][j][k];
          q[numSols][4] = q5[i][j];
	  q[numSols][5] = q6[i][j];
          numSols++;
	}
      }
    }
  }
  
  return numSols;
}     
          
          
